//TODO: Tooltips on chart. Especially max/min ones displayed automatically.
(function () {
	'use strict';
	var MINUTES_TO_MILLISECONDS = 1000 * 60,
		MILLISECONDS_TO_MINUTES = 1 / MINUTES_TO_MILLISECONDS,
		HOURS_TO_MILLISECONDS = MINUTES_TO_MILLISECONDS * 60,
		MILLISECONDS_TO_HOURS = 1 / HOURS_TO_MILLISECONDS,
		DAYS_TO_MILLISECONDS = HOURS_TO_MILLISECONDS * 24,
		MILLISECONDS_TO_DAYS = 1 / DAYS_TO_MILLISECONDS,
		PASCALS_TO_KILOPASCALS = (1 / 1000),
		DAY_NAMES = ['Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat'],
		MONTH_NAMES = ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec'],
		currentUpdateSeconds = 10,
		dateCeil = {
			//Returns a Date rounded up to the next whole hour.
			hour: function (date) {
				return new Date(date.getFullYear(), date.getMonth(), date.getDate(), date.getHours() + (date.getMinutes() ? 1 : 0));
			},
			day: function (date) {
				return new Date(date.getFullYear(), date.getMonth(), date.getDate() + ((date.getHours() || date.getMinutes()) ? 1 : 0));
			},
			month: function (date) {
				return new Date(date.getFullYear(), date.getMonth() + ((date.getDate() || date.getHours() || date.getMinutes()) ? 1 : 0));
			}
		},
		dateToMilliseconds = function (date) {
			//Converts a date into a number, as Chartist doesn't have proper date axis support.
			//Subtract time zone so that integers correspond to midnight local time.
			return (date - (date.getTimezoneOffset() * MINUTES_TO_MILLISECONDS));
		},
		millisecondsToDate = function (ms) {
			var date = new Date(ms);
			//Add time zone that was subtracted in dateToAxisUnits to get back to the real time.
			date.setMinutes(date.getMinutes() + date.getTimezoneOffset());
			return date;
		},
		dateToMonths = function (date) {
			//Months have to be handled specially because they are of variable length.
			var year = date.getFullYear(),
				month = date.getMonth(),
				dayOfMonth = date.getDate(),
				daysInMonth = (new Date(year, month + 1, 0)).getDate(),
				wholeMonths = (year * 12) + month,
				fractionalMonths = (((date - new Date(year, month)) * MILLISECONDS_TO_DAYS) - 1) / daysInMonth;

			return wholeMonths + fractionalMonths;
		},
		monthsToDate = function (value) {
			var totalMonths = Math.floor(value),
				month = totalMonths % 12,
				year = (totalMonths - month) / 12,
				date = new Date(year, month),
				daysInMonth = (new Date(year, month + 1, 0)).getDate(),
				days = (value - totalMonths) * daysInMonth;

			//The -0 ensures the values are added as numbers, not strings.
			return new Date(date - 0 + ((days + 1) * DAYS_TO_MILLISECONDS));
		},
		dayOfMonthToOrdinal = function (dayOfMonth) {
			switch ([11,12,13].indexOf(dayOfMonth) >= 0 ? 0 : dayOfMonth % 10) {
				case 1: return dayOfMonth + 'st';
				case 2: return dayOfMonth + 'nd';
				case 3: return dayOfMonth + 'rd';
				default: return dayOfMonth + 'th';
			}
		},
		periods = {
			day: {
				chartUpdateMinutes: 10,
				parameters: function () {
					var toDate = dateCeil.hour(new Date()),
						fromDate = new Date(toDate.getFullYear(), toDate.getMonth(), toDate.getDate(), toDate.getHours() - 24);
					return {
						toDate: toDate,
						fromDate: fromDate,
						groupBy: ''
					};
				},
				labelInterpolationFnc: function (value) {
					var date = millisecondsToDate(value * HOURS_TO_MILLISECONDS);
					return date.getHours() + ':00';
				},
				dateToAxisUnits: function (date) {
					return dateToMilliseconds(date) * MILLISECONDS_TO_HOURS;
				},
				axisUnitsToDate: function (value) {
					return millisecondsToDate(value * MILLISECONDS_TO_HOURS);
				},
				isMajorGrid: function (value) {
					var date = millisecondsToDate(value * HOURS_TO_MILLISECONDS);
					return date.getHours() % 6 === 0;
				}
			},
			week: {
				chartUpdateMinutes: 30,
				parameters: function () {
					var toDate = dateCeil.day(new Date()),
						fromDate = new Date(toDate.getFullYear(), toDate.getMonth(), toDate.getDate() - 7);
					return {
						toDate: toDate,
						fromDate: fromDate,
						groupBy: 'hour'
					};
				},
				labelInterpolationFnc: function (value) {
					var date = millisecondsToDate(value * DAYS_TO_MILLISECONDS);
					return DAY_NAMES[date.getDay()];
				},
				dateToAxisUnits: function (date) {
					return dateToMilliseconds(date) * MILLISECONDS_TO_DAYS;
				},
				axisUnitsToDate: function (value) {
					return millisecondsToDate(value * MILLISECONDS_TO_DAYS);
				},
				isMajorGrid: function (value) {
					var date = millisecondsToDate(value * DAYS_TO_MILLISECONDS),
						day = date.getDay();
					return day === 1 || day === 6;
				}
			},
			month: {
				chartUpdateMinutes: 180,
				parameters: function () {
					var toDate = dateCeil.day(new Date()),
						//Always go back 31 days rather than the length of the current month, just because it's simpler.
						fromDate = new Date(toDate.getFullYear(), toDate.getMonth(), toDate.getDate() - 31);
					return {
						toDate: toDate,
						fromDate: fromDate,
						groupBy: 'hour'
					};
				},
				labelInterpolationFnc: function (value) {
					var date = millisecondsToDate(value * DAYS_TO_MILLISECONDS);
					return dayOfMonthToOrdinal(date.getDate());
				},
				dateToAxisUnits: function (date) {
					return dateToMilliseconds(date) * MILLISECONDS_TO_DAYS;
				},
				axisUnitsToDate: function (value) {
					return millisecondsToDate(value * MILLISECONDS_TO_DAYS);
				},
				isMajorGrid: function (value) {
					var date = millisecondsToDate(value * DAYS_TO_MILLISECONDS);
					return date.getDate() === 1;
				}
			},
			year: {
				chartUpdateMinutes: 720,
				parameters: function () {
					var toDate = dateCeil.month(new Date()),
						fromDate = new Date(toDate.getFullYear(), toDate.getMonth()- 12);
					return {
						toDate: toDate,
						fromDate: fromDate,
						groupBy: 'day'
					};
				},
				labelInterpolationFnc: function (value) {
					var date = monthsToDate(value);
					return MONTH_NAMES[date.getMonth()];
				},
				dateToAxisUnits: function (date) {
					return dateToMonths(date);
				},
				axisUnitsToDate: function (value) {
					return monthsToDate(value);
				},
				isMajorGrid: function (value) {
					var date = monthsToDate(value);
					return date.getMonth() % 3 === 0;
				}
			}
		},
		selectedPeriod = function () { return periods[document.querySelector('#period input:checked').value]; },
		ajax = function (method, url) {
			var promise = new Promise(function (resolve, reject) {
				var httpRequest = new XMLHttpRequest();
				httpRequest.open(method, url);
				httpRequest.send();
				httpRequest.onload  = function () {
					if (this.status >= 200 && this.status < 300) {
						resolve(this.responseText.length ? JSON.parse(this.responseText) : this.responseText);
					} else {
						reject(this.statusText);
					}
				};
				httpRequest.onerror = function () {
				  reject(this.statusText);
				};
			});
			return promise;
		},
		chartOptions = function (limits) {
			return {
				axisX: {
					type: Chartist.AutoScaleAxis,
					onlyInteger: true,
					scaleMinSpace: 15,
					high: limits ? limits.high : undefined,
					low: limits ? limits.low : undefined,
					labelInterpolationFnc: selectedPeriod().labelInterpolationFnc
				},
				axisY: {
					type: Chartist.AutoScaleAxis,
					onlyInteger: true,
					scaleMinSpace: 15,
					offset: 30
				},
				showPoint: false,
				lineSmooth: false,
				chartPadding: { top: 10, right: 15, bottom: 5, left: 0 }
			};
		},
		chartOptionsResponsive = function () {
			return [
				['(max-width: 50rem)', { chartPadding: { right: 10 }, axisY: { offset: 28 } }],
				['(max-width: 30rem)', { chartPadding: { top: 3, right: 5, bottom: 0 }, axisX: { offset: 25 }, axisY: { offset: 25 } }]
			];
		},
		locations,
		measures = [
			{
				name: 'temperature',
				label: 'Temperature',
				unit: '°C',
				dataTransform: function (temperature) { return temperature; },
				chartOptions: function (readings, limits) {
					readings = readings || [];
					return Chartist.extend({}, chartOptions(limits), {
						axisY: {
							low: Math.floor(readings.reduce(function (min, reading) { return Math.min(reading.temperature, min); }, 0)),
							high: Math.ceil(readings.reduce(function (max, reading) { return Math.max(reading.temperature, max); }, 25))
						}
					});
				},
				isMajorGrid: function (value, bounds) {
					return value % 5 === 0;
				},
				chart: undefined,
				data: []
			},
			{
				name: 'pressure',
				label: 'Pressure',
				unit: 'kPa',
				dataTransform: function (pressure) { return pressure * PASCALS_TO_KILOPASCALS; },
				chartOptions: function (readings, limits) {
					readings = readings || [];
					return Chartist.extend({}, chartOptions(limits), {
						axisY: {
							low: Math.floor(readings.reduce(function (min, reading) { return Math.min(reading.pressure, min); }, 98)),
							high: Math.ceil(readings.reduce(function (max, reading) { return Math.max(reading.pressure, max); }, 103))
						}
					});
				},
				isMajorGrid: function (value, bounds) {
					return value % 5 === 0;
				},
				chart: undefined,
				data: []
			},
			{
				name: 'humidity',
				label: 'Humidity',
				unit: '%',
				dataTransform: function (humidity) { return humidity; },
				chartOptions: function (readings, limits) {
					return Chartist.extend({}, chartOptions(limits), {
						axisY: {
							low: 0,
							high: 100
						}
					});
				},
				isMajorGrid: function (value, bounds) {
					var step = bounds.step >= 10 ? 50 : 10;
					return value % step === 0;
				},
				chart: undefined,
				data: []
			},
		],
		getLocations = function () {
			return ajax('GET', '/locations/');
		},
		getCurrentReadings = function (location) {
			return ajax('GET', '/readings/' + location);
		},
		getDateReadings = function (location, fromDate, toDate, groupBy) {
			var url = '/readings/' + location + '?fromDate=' + fromDate.toISOString() + '&toDate=' + toDate.toISOString() + '&groupBy=' + groupBy;
			return ajax('GET', url);
		},
		initMeasures = function () {
			var container = document.querySelector('#container'),
				template = document.querySelector('#template-measure');
			measures.forEach(function (measure) {
				var measureContainer = document.importNode(template.content.children[0], true),
					title = measureContainer.querySelector('.title');
				measureContainer.classList.add('measure');
				measureContainer.classList.add(measure.name);
				title.appendChild(document.createTextNode(measure.label + ' / ' + measure.unit));
				container.appendChild(measureContainer);
			});
		},
		initCurrentReadings = function () {
			var template = document.querySelector('#template-current');
			//Locations are added above each other. Iterate in reverse so the first one ends up on top.
			locations.reverse().forEach(function (location, index) {
				measures.forEach(function (measure) {
					var container = document.querySelector('#container > .' + measure.name),
						current = document.importNode(template.content.children[0], true),
						label = current.querySelector('.label');
					current.classList.add('location' + index);
					label.appendChild(document.createTextNode(location));
					container.insertBefore(current, container.firstElementChild.nextElementSibling);
				});
			});
		},
		initCharts = function () {
			measures.forEach(function (measure) {
				measure.chart = new Chartist.Line('#container .' + measure.name + ' .ct-chart', {}, measure.chartOptions(), chartOptionsResponsive());
				measure.chart.on('draw', function (data) {
					var value;
					if (data.type === 'grid') {
						if (data.axis.units.pos === 'y') {
							value = data.axis.bounds.values[data.index];
							if (measure.isMajorGrid(value, data.axis.bounds)) {
								data.element.addClass('major');
							}
						} else if (data.axis.units.pos === 'x') {
							value = data.axis.bounds.values[data.index];
							if (selectedPeriod().isMajorGrid(value)) {
								data.element.addClass('major');
							}
						}
					}
				});
			});
		},
		updateCurrentReadings = function () {
			return Promise.all(locations.map(function (location, index) {
				return getCurrentReadings(location).then(function (data) {
					measures.forEach(function (measure) {
						var current = document.querySelector('#container .' + measure.name + ' .current.location' + index),
							quantity = current.querySelector('.quantity'),
							value = measure.dataTransform(data[measure.name]).toFixed(1);
						quantity.innerHTML = '';
						quantity.appendChild(document.createTextNode(value));
					});
				});
			}));
		},
		updateCharts = function () {
			var period = selectedPeriod(),
				parameters = period.parameters(),
				limits = {
					low: period.dateToAxisUnits(parameters.fromDate),
					high: period.dateToAxisUnits(parameters.toDate)
				};
			return Promise.all(locations.map(function (location, index) {
				return getDateReadings(location, parameters.fromDate, parameters.toDate, parameters.groupBy).then(function (data) {
					return data.map(function (reading) {
						var ms = Date.parse(reading.date),
							transformedReading = { date: period.dateToAxisUnits(new Date(ms)) };

						measures.forEach(function (measure) {
							transformedReading[measure.name] = measure.dataTransform(reading[measure.name]);
						});

						return transformedReading;
					});
				});
			})).then(function (values) {
				var flattenedValues = values.reduce(function(a, b) {
						return a.concat(b);
					}, []);
				measures.forEach(function (measure) {
					measure.data = values.map(function (transformedReadings, index) {
						return {
							className: 'location' + index + ' ' + measure.name,
							name: location,
							data: transformedReadings.map(function (reading) { return { x: reading.date, y: reading[measure.name] }; })
						};
					});
					measure.chart.update({ series: measure.data }, measure.chartOptions(flattenedValues, limits));
				});
			});
		},
		backoff = function (idealDelay, retryCount, initialRetryDelay) {
			var randomizeByFactor = function (value, factor) {
					return value * (1 + ((factor * 2 * Math.random()) - factor));
				},
				backedOffDelay = retryCount > 0 ? Math.min(idealDelay, initialRetryDelay * Math.pow(2, retryCount - 1)) : idealDelay;
			//Randomize a bit to reduce the chance of multiple clients all updating at once, moreso when backing off.
			return randomizeByFactor(backedOffDelay, retryCount > 0 ? 0.5 : 0.1);
		},
		updateChartsTimerId,
		updateChartsTimer = function (retryCount) {
			var again = function (nextRetryCount) {
				var delay = backoff(selectedPeriod().chartUpdateMinutes * 60, nextRetryCount, 5);
				clearTimeout(updateChartsTimerId);
				updateChartsTimerId = setTimeout(updateChartsTimer, delay * 1000, nextRetryCount);
			};
			return updateCharts().then(function () {
				again(0);
			}).catch(function (error) {
				console.error('Failed to update charts: %o', error);
				again((retryCount || 0) + 1);
			});
		},
		updateCurrentReadingsTimerId,
		updateCurrentReadingsTimer = function (retryCount) {
			var again = function (nextRetryCount) {
				var delay = backoff(currentUpdateSeconds, nextRetryCount, 2);
				clearTimeout(updateCurrentReadingsTimerId);
				updateCurrentReadingsTimerId = setTimeout(updateCurrentReadingsTimer, delay * 1000, nextRetryCount);
			};
			return updateCurrentReadings().then(function () {
				again(0);
			}).catch(function (error) {
				console.error('Failed to update current readings: %o', error);
				again((retryCount || 0) + 1);
			});
		},
		togglePeriodButtons = function (enable) {
			//Could make this tidier with a for-of loop, but it doesn't work in Chrome. See https://bugs.chromium.org/p/chromium/issues/detail?id=538558
			//"slice" converts the NodeList from querySelectorAll into a real array, so that forEach can be used.
			Array.prototype.slice.call(document.querySelectorAll('#period input')).forEach(function (input) {
				if (enable) {
					input.removeAttribute('disabled');
				} else {
					input.setAttribute('disabled', 'disabled');
				}
			});
		};

	if (window.location.hostname === 'localhost') {
		var button = document.getElementById('display');
		button.addEventListener('click', function () {
			ajax('POST', '/display?enable=false').catch(function (error) {
				console.error('Failed to turn off display: %o', error);
			});
		});
		button.classList.add('local');
	}

	//Can't do anything until the list of sensor locations has been retrieved.
	getLocations().then(function (data) {
		locations = data;
		initMeasures();
		initCurrentReadings();
		initCharts();

		updateCurrentReadingsTimer();
		updateChartsTimer();

		document.getElementById('period').addEventListener('change', function (event) {
			var button = event.target.nextElementSibling;
			if (!document.querySelector('#period .loading')) {
				togglePeriodButtons(false);
				button.classList.add('loading');
				updateChartsTimer().then(function () {
					togglePeriodButtons(true);
					button.classList.remove('loading');
				});
			}
		});
	}).catch(function (error) {
		console.error('Initialization failed: %o', error);
	});
})();
